------------------------------------------------------------------------------------
-- Addon:         ProGuide - Guide Handler UI                                                 
-- Author:        Claus Gade Pedersen - Satisfyme                                 
-- File:          GuideValidator.lua
-- Description:   Holds a list of valid content info to validate a guide and
--                the validate function
------------------------------------------------------------------------------------
-- Addon version: 21                                              
-- File Revision: 12                                                 
-- Last Edit:     2011-09-16T21:44:53Z                                                 
-- Last Edit By:  satisfyme                                                   
------------------------------------------------------------------------------------
local L = ProGuideLocals
local errorReport = {}
local Headlines

local GuideValidatTable = {
  ["Required"] = {"Name", "Part",},
  ["Name"] = "string",                       -- The shown name of the guide
  ["Icon"] = "string",                       -- Top left icon. must be a texture path
  ["Race"] = {"string", "table",},           -- What race that can see this guide
  ["Class"] = {"string", "table",},          -- What class that can see this guide
  ["ShowSkillLevel"] = {"number", "table",}, -- Shows skill name and skill level
  ["Part"] = {                               -- Parts in the guide. only one part is shown at the time. The list must be number in line after eachother
    ["Required"] = {"Title", "Step",},
    ["Title"] = "string",       -- Part title to show what part we are in
    ["UseBuyList"] = "boolean", -- Enable Buy List button is set to true
    ["Step"] = {                -- Steps in the guide. All step is shown in the guide objectives
      ["Headlines"] = {
        "Create",
        "Learn",
        "LevelSkill",
        "TrainSkill",
      },
      ["Create"] = {        -- Create number of items
        [0] = {"table",},
        ["Required"] = {"ItemId", "Num",},
        ["ItemId"] = "number", -- The item id of the item that needs to be created
        ["Num"] = "number",    -- The number af that item that needs to be created
        ["ItemUsed"] = {       -- A list of items needed to create 1 of the create item. Only works if UseBuyList is true
          [0] = {"number", "list",}, -- The item id list can be table or a number
          ["Required"] = {"Num",},
          ["Num"] = "number",        -- num of item needed to create 1 item. Only works if item id is a table
          ["Note"] = "string",       -- A note to the item in the buy list
        },
      },
      ["Learn"] = {         -- Learn how to create an item
        [0] = {"table",},
        ["Required"] = {"Skill", "ItemId",},
        ["Skill"] = "number",
        ["ItemId"] = "number",
        ["LearnFrom"] = "number",
      },
      ["LevelSkill"] = {    -- Add level skill to objectives
        [0] = {"table",},
        ["Required"] = {"Skill", "Level",},
        ["Skill"] = "number", -- The Skill Id
        ["Level"] = "number", -- What skill level to reach
        ["Create"] = { -- A list of items to create for leveling the skill
          [0] = {"list",},
          ["Required"] = {"ItemId", "Num",},
          ["ItemId"] = "number", -- the item id
          ["Num"] = "number",    -- num if items to reach the level
          ["ItemUsed"] = {       -- A list of items needed to create 1 of the create item. Only works if UseBuyList is true
            [0] = {"number", "list",}, -- The item id list can be table or a number
            ["Required"] = {"Num",},
            ["Num"] = "number",        -- num of item needed to create 1 item. Only works if item id is a table
            ["BuyFrom"] = "number",    -- An NPC id from where this item can be bought
          },
        },
        ["CreateEndNote"] = "string", -- Note to add after the create item list
      },
      ["Note"] = {  -- Add a note to objectives
        [0] = {"table",},
        ["Required"] = {"text",},
        ["Text"] = "string", -- The text of the note
        ["Location"] = { -- if set it will add a location to the note that can be shown by clicking it
          [0] = "table",
          ["Required"] = {"x", "y", "zone",},
          ["x"] = "number",    -- x cordinate
          ["y"] = "number",    -- y cordinate
          ["zone"] = "string", -- zone name
        },
      },
      ["Skill"] = { -- A skill checker that will complete when a skill level is reached
        [0] = {"table",},
        ["Required"] = {"Skill", "Level",},
        ["Skill"] = "number", -- The skill id
        ["Level"] = "number", -- Skill level to compleate at
      },
      ["TrainSkill"] = { -- Train a skill to a higher rank
        [0] = {"table",},
        ["Required"] = {"Skill", "Level",},
        ["Skill"] = "number", -- The skill id
        ["Level"] = "number", -- The max skill level of the new rank
        ["TrainFrom"] = "number", -- If set then shows the NPC and waypoint to the NPC where this skill can be trained.
      },
    },
  },
}

-- Set Prefix And Surfix
local function SetPreAndSurfix(key, Part, Step, StepInfo)
  local prefix = ""
  local surfix = ""
  local nokeysurfix = ""

  if not StepInfo then
    surfix = key
    nokeysurfix = key
  else
    surfix = StepInfo .. "->" .. key
    nokeysurfix = StepInfo
  end
  if not Part then
    prefix = L["DevValiPrefixMain"]
  else
    if Step then
      prefix = L["DevValiPrefixStep"]:format(Part, Step, surfix)
    else
      prefix = L["DevValiPrefixPart"]:format(Part)
    end
  end

  return prefix, surfix, nokeysurfix
end

-- Validate Vontent of guide
function ProGuide:ValidateContent(guide, GV, Part, Step, StepInfo)
  -- Locals we need
  local prefix = ""
  local surfix = ""
  local nokeysurfix = ""
  local required = 0

  -- populate the GV (Guide validate table) if not set
  if not GV then GV = GuideValidatTable end

  -- Reset error report if it's the main line
  if not Part then errorReport = {} end

  -- loop throu guide section
  for key,val in pairs(guide) do
    -- set error prefix so we know where we are in the error report
    prefix, surfix, nokeysurfix = SetPreAndSurfix(key, Part, Step, StepInfo)

    -- check if key is in the GuideValidate
    if not GV[key] then
      tinsert(errorReport, L["DevKeyNotValid"]:format(prefix, surfix))
      key = nil
    end

    -- If key is Step then handle steps
    if key == "Step" and not Step then
      for step, info in ipairs(guide.Step) do
        Headlines = 0
        guide.Step[step] = ProGuide:ValidateContent(info, GV.Step, Part, step)

        if Headlines > 1 then
          tinsert(errorReport, L["DevStepOnlyOneHeadline"]:format(L["DevValiPrefixStep"]:format(Part, step), Headlines))
        end
      end

    -- If key is Part then handle parts
    elseif key == "Part" and not Part then
      for part, info in ipairs(guide.Part) do
        guide.Part[part] = ProGuide:ValidateContent(info, GV.Part, part)
      end

    -- Check other key elements
    else
      if key == "Required" then
        guide[key] = nil
        key = nil
      elseif key == "Headlines" then
        guide[key] = nil
        key = nil
      else
        if type(GV[key]) == "table" then
          if GV[key][0] then
            if tContains(GV[key][0], "table") then
              guide[key] = ProGuide:ValidateContent(val, GV[key], Part, Step, surfix)
            elseif tContains(GV[key][0], "list") then
              for num, info in pairs(val) do
                if type(num) ~= "number" then
                  tinsert(errorReport, L["DevKeyMustBeNumber"]:format(prefix, surfix .. "->" .. num))
                elseif num ~= 0 then
                  if type(info) == "table" then
                    guide[key][num] = ProGuide:ValidateContent(info, GV[key], Part, Step, surfix .. "->" .. num)
                  elseif not tContains(GV[key][0], type(info)) then
                    tinsert(errorReport, L["DevKeyValueWrongFormat"]:format(prefix, surfix .. "->" .. num))
                  end
                end
              end
              --guide[key] = ProGuide:ValidateContent(val, GV[key], Part, Step, surfix)
            end
          elseif not tContains(GV[key], type(val)) then
            tinsert(errorReport, L["DevKeyValueWrongFormat"]:format(prefix, surfix))
          end
        elseif type(GV[key]) == "string" and GV[key] ~= type(val) then
          tinsert(errorReport, L["DevKeyValueWrongFormat"]:format(prefix, surfix))
        end

        if not StepInfo and tContains(GuideValidatTable.Part.Step.Headlines, key) then
          Headlines = Headlines + 1
        end
      end
    end

    -- check if it is a required field
    if GV.Required and key and tContains(GV.Required, key) then required = required + 1 end
  end

  -- if we dont have all the required fields then return as a failed guide
  if GV.Required and #GV.Required ~= required then tinsert(errorReport, L["DevRequiredMissing"]:format(#GV.Required - required, prefix, nokeysurfix)) end

  -- return validated guide/part/step
  if Part then
    return guide
  else
    return guide, errorReport
  end
end


--[[
  if isStep then
    return guide
  elseif isPart then
    for part, info in ipairs(guide) do
      required = 0
      for key, val in pairs(info) do
        -- check if key is in the GuideValidate
        if not GV.Part[key] then
          tinsert(errorReport, L["|cffff0000Part %d, Key:|r |cff00ff00%s|r|cffff0000, is not a valid key.|r"]:format(part, key))
          key = nil
        end

        if key then
          -- If key is Step then handle steps
          if key == "Step" then
            guide[part][key] = ValidateContent(val, part, true)

          -- check if key type matches the validator
          else
            if key == "Required" then
              guide[part][key] = nil
              key = nil
            elseif
                (type(GV.Part[key]) == "table" and not tContains(GV.Part[key], type(val)))
                or (type(GV.Part[key]) == "string" and GV.Part[key] ~= type(val))
              then
                tinsert(errorReport, L["|cffff0000Part %d, Key:|r |cff00ff00%s|r|cffff0000 - value, is wrong format.|r"]:format(part, key))
            end
          end
        end

        -- check if it is a required field
        if key and tContains(GV.Part.Required, key) then required = required + 1 end
      end
      if #GV.Part.Required ~= required then tinsert(errorReport, L["|cffff6060%d required key[s] is missing in Part %d.|r"]:format(#GV.Part.Required - required, part)) end
    end

    return guide
  else
 ]]

--[[
      ["Headlines"] = {
        ["Boat"] = {
          ["Required"] = {"Location",},
          ["Location"] = "string",
        },
        ["Fly"] = {
          ["Required"] = {"Location",},
          ["Location"] = "string",
        },
        ["GetFlighgtPoint"] = {
          ["Required"] = {"FlighgtPoint", "x", "y", "zone",},
          ["FlighgtPoint"] = "string",
          ["x"] = "number",
          ["y"] = "number",
          ["zone"] = "string",
        },
        ["QuestAccept"] = {
          ["Required"] = {"Qid", "QuestName",},
          ["Qid"] = "number",
          ["QuestName"] = "string",
        },
        ["QuestObjective"] = {
          ["Required"] = {"Qid", "QuestName",},
          ["Qid"] = "number",
          ["QuestName"] = "string",
        },
        ["QuestTurnIn"] = {
          ["Required"] = {"Qid", "QuestName",},
          ["Qid"] = "number",
          ["QuestName"] = "string",
        },
        ["Route"] = {
          ["Required"] = {"Text", "Waypoints",},
          ["Text"] = "string",
          ["Waypoints"] = {
            "list",
            ["Required"] = {"x", "y", "zone",},
            ["x"] = "number",
            ["y"] = "number",
            ["zone"] = "string",
          },
        },
        ["Teleport"] = {
          ["Required"] = {"Location",},
          ["Location"] = "string",
        },
      },
      ["SubHeadlines"] = {
        ["Note"] = {
          ["Required"] = {"text",},
          ["Text"] = "string",
          ["Location"] = {
            ["Required"] = {"x", "y", "zone",},
            ["x"] = "number",
            ["y"] = "number",
            ["zone"] = "string",
          },
        },
        ["Skill"] = {
          ["Required"] = {"Skill", "Level",},
          ["Skill"] = "number",
          ["Level"] = "number",
        },
      },
 ]]

 ------------------------------------------------------------------------------------
-- More detailed info can be found here:
-- http://github.com/tekkub/TourGuide/blob/docs/writing_guides.markdown
------------------------------------------------------------------------------------
-- Headlines and Objective icons
------------------------------------------------------------------------------------
-- A QuestName |QID|12593| ---------- Accept quest
-- C QuestName |QID|8325| ----------- Kill stuff (?)
-- T QuestName |N|note| |QID|12593| - Turnin quest
-- b Stormwind City ----------------- Boat to Stormwind City
-- f Stormwind City ----------------- Get flight point
-- F Ironforge ---------------------- Fly to Ironforge
-- h Falconwing Square -------------- Set hearth loc to Falconwing Square
-- H Falconwing Square -------------- Use hearth to go to loc Falconwing Square
-- R Dawning Lane ------------------- Go to Dawning Lane
-- N Kill any cats you come across -- Note entry
-- B [Suntouched Special Reserve]---- Buy something
-- U Orb of Returning |U|24335| ----- Use item
-- K Nightsaber --------------------- Kill stuff entry(?)
--'
------------------------------------------------------------------------------------
-- Sub objectives
------------------------------------------------------------------------------------
-- |N|Somethins is wrong here| ---------------- Note: Note shown when mouse over objective frame
-- |N|Somtehins is wrong here (34.43, 53.23)| - Note: Same as abow, just with a cordinate that will be send to TomTom or Cartographer
-- |QID|458| ---------------------------------- Quest ID: Used with A, C, T Headlines to tell witch quest it is.
-- |O| ---------------------------------------- Optional: Tells that this quest is optional
-- |L|1234| ----------------------------------- Loot Item ID: Complete when you loot an item
-- |L|1234 10| -------------------------------- Loot Item ID: Same as abowe, just that it will first complete when there have been looted 10 of the item
-- |U|1234| ----------------------------------- Use Item ID: Pops up a button that you can use to use the specifik item
-- |PRE|Quest Name| |O| ----------------------- Prerequsite: Prerequsite quest that need to be done before showing. NOTE: must be optional as well
-- |C|Priest,..| ------------------------------ Class Limit: Objective is limited to one or more classes. multiclass is splited with a ","
--    Classes: Death Knight, Druid, Hunter, Mage, Paladin, Priest, Rogue, Shaman, Warlock, Warrior
-- |R|Human,..| ------------------------------- Race Limit: Objective is limited to one or more races. multirace is splited with a ","
--    Alliance Races: Human, Dwarf, Night Elf, Gnome, Draenei, Worgen
--    Horde Races: Orc, Undead, Tauren, Troll, Blood Elf, Goblin
-- |Z|Darnassus| ------------------------------ Zone: Used to specifi another zone is the objective is on another zone
-- |Q|Quest Name| |QO|Some Mob slain: 10/10| -- Sub-objective: must match the quest log text EXACTLY. Completes when a sub-objective of the quest matches the QO tag. Both values must be specified.
-- |T| ---------------------------------------- Town: Signifies that an objective takes place in a "town". Will only change background color of queue
--'
------------------------------------------------------------------------------------
-- Added functions (Only works if Tour Guide - Trade Skill addon is installed)
------------------------------------------------------------------------------------
-- |CURRENCY|id amount| ---------------- Completes when you have the needed amount of currency.
--    Known id: 61  - Dalaran Jewelcrafting Token
--              81  - Dalaran Cooking Award
--              241 - Champion's Seal
--              361 - Illustrious Jewelcrafter's Token
--              390 - Conquest Points
--              391 - Tol Barad Commendation
--              392 - Honor Points
--              395 - Justice Points
--              396 - Valor Points
--              402 - Chef's Award
-- |H|The Commons| --------------------- Completes if Heartstone is set to that location
--    To be sure it's the right binding location, first bind to that location and check the system message
--    If you already are bind to that location, then you can use the slash command: /heartloc to get the location
-- |LEARNITEM|Filet of Redgill| -------- Completes when you learned to create that item
--    This tag is also used for Auto Learn From Trainer in the option panel.
-- |LOOT|1234 20| ---------------------- Same as |L| but with a corrected bug + it will add an counter to show how far you are in the progress.
-- |LOOT|1234 20, 5678 15, ...| -------- Same as |LOOT| but it will now check for more than just 1 item and all items is listed in the little window over the objective frame.
--    The max number of items to be listed is 6
--    This tag is also used for Auto Buy From Merchant in the option panel.
-- |LVL|43| ---------------------------- Level requirement: Complete when a level exact or higher that specified.
-- |LVLTOSKILL|Fishing 2| -------------- Same as |SKILL| but cant add an second Skillname. It will also add a counter to show how far you are.
-- |MAP|MapFileName[, ZoneName[, MinLevel[, MaxLevel]]]| Show an map icon. When clicked, a map will popup. When the map is clicked it will be larger.
--    MapFileName:
--       MapFileName is the file name of the map stored in "Interface\AddOns\TourGuide-Tradeskill\Maps" folder without the .tga file ending
--       MapFileName must be of imagetype tga.
--       The tga dimension must be 512x256.
--       The tga will be clipped in wow so the usable size is 384x256 left aligned.
--       A tga example can be found in "Interface\AddOns\TourGuide-Tradeskill\Maps" folder.
--       A Photoshop example can be found in "Interface\AddOns\TourGuide-Tradeskill\Maps" folder.
--       This site have maps you can use: http://www.wowhead.com/maps
--       Select what map to use and right click it and then save background image or show background image and then save it.
--    ZoneName: (Optional)
--      ZoneName is the name of the zone, where the map is from.
--      When added, a arrow icon will be shown on the map at the current player position and the direction the player is heading. The arrow will only show if you are in the correct zone.
--    MinLevel: (Optional)
--      MinLevel is the minimum level required to see this map.
--    MaxLevel: (Optional)
--      MaxLevel is the maximum level required to see this map.
--    MinLevel and MaxLevel can be handy if you want to show different maps depending on what level the player are.
-- |SKILL|Fishing 2, ...|--------------- Skillname and Rank: NOTE: case-sensitive. Complete if skill rank is reached.
--    When given a second Skillname and Rank (Rank is optional), then both rank needs to be reached before completing
-- |HAVESKILL|First Aid 200| ----------- Skillname and Rank: NOTE: case-sensitive. Show objective if you have the skill your rank is equal or higher that required.
-- |DONTHAVESKILL|First Aid 200| ------- Skillname and Rank: NOTE: case-sensitive. Show objective if you don't have the skill or your rank is lower that required.
-- |TLOOT|ItemId1 ItemId2 ..., Amount| - Used to get a total amount from 2 - 6 items in a list shown above the objective frame.
--    Completes when a total amount of all 2-6 items together reach the amount added in the TLOOT tag.
--    Mostly used to let the player chose what to buy/get
--    Space is used to separate the itemID's and the comma is used to separate the itemID's from the Amount needed.
-- |TRAINSKILL|Fishing 75| ------------- Skillname and MaxRank: NOTE: case-sensitive. Complete when training a skill. The number must be the max rank that the skill will be after training
--    This tag is also used for Auto Train From Trainer in the option panel.
--'
------------------------------------------------------------------------------------
-- Added slash commands (Only works if Tour Guide - Trade Skill addon is installed)
------------------------------------------------------------------------------------
-- /itemid [Name of item] - Show item id in chat window. Use: open chatbox, type in /itemid and the shift + click on that item to get item id from and press enter.
-- /skillinfo ------------- Show all your skills in chat window. will show all skills both trade skills and non-trade skills
-- /heartloc -------------- Show your current Heartstone binding location